#!/usr/bin/env python3

from argparse import ArgumentParser
from itertools import combinations
from sys import stdout
import networkx as nx
import numpy as np


def readMatrix(data):

    res = list()
    for line in data:
        res.append(list(map(int, line[:-1])))
    return np.array(res, dtype=bool)


def constructCompatiblityMatrix(M):
    """ constructs only upper triangle matrix """

    res = np.zeros(shape = (M.shape[1], M.shape[1]), dtype=bool)
    for i, j in combinations(range(M.shape[1]), 2):
        types = set((M[k, i], M[k, j]) for k in range(M.shape[0]) if M[k,i] or
                M[k, j])
        res[i, j] = len(types) < 3
    return res


def writeILP(G, out):

    # objective 
    print('maximize k', file=out)

    # constraints
    print('subject to', file=out)
    # constraint: k must be equal to the number of active vertices
    print('{} - k = 0'.format(' + '.join('y{}'.format(v) for v in G.nodes())),
            file=out)

    # constraint: any node that has an active edge must be active, too
    for u in G.nodes():
        for v in G.neighbors(u):
            print('x{}_{} - y{} <= 0'.format(*(sorted((u,v)) + [u])), file=out)

    # constraint: any active node must have at least k active edges
    n = G.number_of_nodes()
    for u in G.nodes():
        left_sum = ''
        if G.degree(u):
            left_sum = ' - ' + (' - '.join('x{0}_{1}'.format(*sorted((u, v)))
                for v in G.neighbors(u)))
        print('k {} + {}y{} <= {}'.format(left_sum, n, u, n+1), file=out)

    # declaration of binary variables
    print('binary', file=out)
    for u, v in G.edges():
        print('x{}_{}'.format(*sorted((u, v))), file=out)
    for v in G.nodes():
        print('y{}'.format(v), file=out)
    print('end', file=out)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('binary_matrix', type=open, help='zero/one matrix')
    args = parser.parse_args()

    # read character matrix
    M = readMatrix(args.binary_matrix)
    # character compatibility matrix
    C = constructCompatiblityMatrix(M)
    # graph with vertices representing characters; any two vertices share an
    # edge if their corresponding characters are compatible
    G = nx.from_numpy_matrix(C)
    # construct & output ILP
    writeILP(G, stdout)
